package monitor;

/** A condition queue. Uses the signal and wait discipline (SW)
    @see Monitor
*/
public class Condition
{
    private Monitor home_monitor ;
    private Assertion prop ;
    private BinarySemaphore queue ;
    private int count ;

    Condition( Monitor _home_monitor, Assertion _prop ) {
        home_monitor = _home_monitor ;
        prop = _prop ;
        queue = new BinarySemaphore(true) ;
        count = 0 ; }

    public Assertion get_assertion() {
        return prop ; }

    /** Wait until a condition is signaled.
        The thread waits outside the monitor until the
        condition is signalled.
        <p> Precondition: Increasing the count by 1
            must make the invariant true.
            This thread is in the monitor.
        <p> Postcontition: The proposition associated
            with this condition queue. This thread is in the monitor.
    */
    public void condition_wait() {
        count += 1 ;
        home_monitor.invariant.test() ;
        home_monitor.entrance.release() ;
        queue.acquire() ;
        count -= 1 ;
        prop.test() ;
    }

    /** Wait only if the condition is not already true */
    public void conditional_condition_wait() {
        if( ! prop.assertion() ) condition_wait() ; }

    /** Signal a condition. Allows one thread that was
        waiting on the conidition to reenter the monitor.
        Consequently the sigalling thread waits outside.
        <p> Preconditions:
            <ul><li>If empty(), the monitor's invariant must be true.</li>
                <li>If not empty(), then decreasing count() by 1 must make
                the proposition associated with this condition true.</li>
                <li>This thread is in the monitor.</li>
            </ul>
        <p> Postcondition: The monitor's invariant.
        This thread is in the monitor.
    */
    public void condition_signal() {

        if( count > 0 ) {
            queue.release() ;
            home_monitor.entrance.acquire() ; }
        else {
            home_monitor.invariant.test() ;
            home_monitor.entrance.release() ;
            home_monitor.entrance.acquire() ; }

        home_monitor.invariant.test() ;
    }

    /** Signal a condition conditionally.
     * Equivalent to if( cond.get_assertion().assertion() ) cond.conditionSignal() ;
    */
    public void conditional_condition_signal() {
        if( get_assertion().assertion() ) condition_signal() ;
    }

    /** Signal a condition. Allows one thread that was
        waiting on the condition to reenter the monitor.
        This thread leaves the monitor.
        <p> Preconditions:
            <ul><li>If empty(), the monitor's invariant must be true.</li>
                <li>If not empty(), then decreasing count() by 1 must make
                the proposition associated with this condition true.</li>
                <li>This thread is in the monitor.</li>
            </ul>
        <p> Postcondition: This thread is not in the monitor.
    */
    public void condition_signal_and_leave()  {
        if( count > 0 ) {
            queue.release() ; }
        else {
            home_monitor.invariant.test() ;
            home_monitor.entrance.release() ; } }


    /** Signal a condition conditionally.
     * Equivalent to if( cond.get_assertion().assertion() ) cond.condition_signal_and_leave() ;
    */
    public void conditional_condition_signal_and_leave() {
        if( get_assertion().assertion() ) condition_signal_and_leave() ;
    }

    /** Test if any thread is waiting on this condition.
        @return true if no thread is waiting; false otherwise.
    */

    public boolean empty() { return count == 0 ; }

    /** How many thread are waiting on this condition.
        @return true if no thread is waiting; false otherwise.
    */
    public int count() { return count ; }

}