package monitor;

/** A class for simulating Monitors. Monitors provide
    coordination of concurrent threads. Each Monitor protects
    some resource, usually data.  At each point in time a monitor
    object is occupied by at most one thread.
    <p> See <br>C.A.R. Hoare, 'Monitors: an operating system structuring
        concept', <EM> Communications of the ACM 17</EM>, Oct. 1974,
        pp. 549-557. Reprinted in C.A.R.Hoare and C.B. Jones,
        <EM> Essay's in Computing Science</EM>, Prentice-Hall, 1989.
        <br>and<br> P. Brinch Hansen, The programming language
        Concurrent Pascal, <EM>IEEE Transactions on Software
        Engineering 1</EM>, June 1975, 199-207.
    <p> The details of pre- and postconditions for signal and wait
        and the idea to use a length function come from Ole-Johan Dahl,
        'A Note on Monitor Versions', http://heim.ifi.uio.no/~olejohan/, 1999.
    @see Condition
*/
public class Monitor
{
    BinarySemaphore entrance = new BinarySemaphore( false ) ;
    Assertion invariant ;

    /** Create a Monitor. */
    public Monitor() {
        invariant = True_Assertion.singleton ; }

    /** Set the invariant.
     *  The default is "true". */
    public void setInvariant( Assertion inv ) { invariant = inv ; }

    /** Enter the monitor. Any thread calling this method is
        delayed until the monitor is unoccupied. Upon returning
        from this method, the monitor is considered occupied.
        <p> A thread must not attempt to enter a Monitor it is
        already in.
    */
    public void enter() {
        entrance.acquire() ;
        invariant.test() ; }

    /** Leave the monitor. After returing from this method, the
    thread no longer occupies the monitor.
    <p> Only a thread that is in the monitor may leave it.
    */
    public void leave() {
        invariant.test() ;
        entrance.release(); }

    /** Create a condition queue associated with a checked Assertion.
        The Assertion will be checked prior to an signal of the
        condition.
    */
    public Condition make_condition( Assertion prop ) {
        return new Condition( this, prop ) ; }

    /** Create a condition queue with no associated checked Assertion.
    */
    public Condition make_condition() {
        return make_condition( True_Assertion.singleton ) ; }
}